from ellipsis.util.root import reprojectRaster
from ellipsis.util.root import reprojectVector

from ellipsis.util.root import saveRaster

from ellipsis.util.root import parseGlb

from ellipsis.util.root import plotRaster
from ellipsis.util.root import plotVector
from ellipsis.util.root import plotPointCloud

from ellipsis.util.root import chunks
from ellipsis.util.root import cover
from ellipsis.util.root import loadingBar

from ellipsis.util.root import transformPoint

from ellipsis.util.root import stringToDate
from ellipsis.util.root import dateToString

from ellipsis.util.root import cutIntoTiles

from ellipsis.util.root import mergeGeometriesByColumn
from ellipsis.util.root import simplifyGeometries



from ellipsis.path.vector.timestamp.feature.root import add, trash, recover, edit, versions
from ellipsis.path.vector.timestamp.feature import message
from ellipsis.path.vector.timestamp.feature import series




#python3 setup.py sdist bdist_wheel
#twine upload --repository pypi dist/*

from ellipsis import account
from ellipsis import user
from ellipsis import path
from ellipsis import view
from ellipsis import util

__version__ = '3.1.28'



from ellipsis import sanitize
from ellipsis.util.root import getActualExtent
from ellipsis import apiManager
from ellipsis.util import loadingBar
from ellipsis.util.root import reprojectRaster
from ellipsis.path.raster.timestamp.util import constructImage
from ellipsis.path.raster.timestamp.util import cutOfTilesPerZoom

from rasterio.io import MemoryFile
from skimage.transform import resize
import json
import numpy as np
from io import BytesIO
import rasterio
import math
import tifffile
from PIL import Image
import geopandas as gpd
import datetime
import pandas as pd
import time
import requests
from skimage.measure import find_contours
from shapely.geometry import Point, LineString

def getDownsampledRaster(pathId, timestampId, extent, width, height, epsg=3857, style = None, token = None):
    return getSampledRaster(pathId, timestampId, extent, width, height, epsg, style, token)

def getSampledRaster(pathId, timestampId, extent, width, height, epsg=4326, style = None, token = None):
    bounds = extent
    token = sanitize.validString('token', token, False)
    pathId = sanitize.validUuid('pathId', pathId, True)
    timestampId = sanitize.validUuid('timestampId', timestampId, True)
    bounds = sanitize.validBounds('bounds', bounds, True)
    style = sanitize.validObject('style', style, False)
    epsg = sanitize.validInt('epsg', epsg, True)
    body = {'pathId':pathId, 'timestampId':timestampId, 'extent':bounds, 'width':width, 'height':height, 'style':style, 'epsg':epsg}
    r = apiManager.get('/path/' + pathId + '/raster/timestamp/' + timestampId + '/rasterByExtent', body, token, crash = True, parseJson = False)


    if type(style) == type(None):
        r = tifffile.imread(BytesIO(r.content))
    else:
        r = np.array(Image.open(BytesIO(r.content)))
    #tif also has bands in last channel
    r = np.transpose(r, [2,0,1])

    xMin = bounds['xMin']
    yMin = bounds['yMin']
    xMax = bounds['xMax']
    yMax = bounds['yMax']



    trans = rasterio.transform.from_bounds(xMin, yMin, xMax, yMax, r.shape[2], r.shape[1])

    return {'raster': r, 'transform':trans, 'extent': {'xMin' : xMin, 'yMin': yMin, 'xMax': xMax, 'yMax': yMax}, 'crs':"EPSG:" + str(epsg) }


def contour(pathId, timestampId, extent, interval = None, intervals = None, epsg = 4326, bandNumber = 1, token = None):
    pathId = sanitize.validUuid('pathId', pathId, True)
    timestampId = sanitize.validUuid('timestampId', timestampId, True)
    token = sanitize.validString('token', token, False)
    epsg = sanitize.validInt('epsg', epsg, True)
    extent = sanitize.validBounds('extent', extent, True)


    res = getActualExtent(extent['xMin'], extent['xMax'], extent['yMin'], extent['yMax'], 'EPSG:' + str(epsg))    
    if res['status'] == '400':
        raise ValueError('Invalid epsg and extent combination')
        
    bounds = res['message']

    xMinWeb = bounds['xMin']
    yMinWeb = bounds['yMin']
    xMaxWeb = bounds['xMax']
    yMaxWeb = bounds['yMax']

    extentWeb = {'xMin': xMinWeb, 'xMax':xMaxWeb, 'yMin':yMinWeb, 'yMax':yMaxWeb}
    size = 1000
    r = getSampledRaster(pathId = pathId, timestampId = timestampId, extent = extentWeb, width = size, height = size, epsg=3857, token = token)
    raster = r['raster']
    
    if bandNumber >= raster.shape[0]:
        raise ValueError('BandNumber too high. Raster only has ' + str(raster.shape[0]) + 'bands.')


    
    if type(intervals) == type(None):
        minVal = np.min(raster[bandNumber-1, raster[-1,:,:] == 1])
        maxVal = np.max(raster[bandNumber-1, raster[-1,:,:] == 1])
        
        
        cont = minVal + interval

        conts = []
        while cont < maxVal:
            conts = conts + [cont]
            cont = cont + interval
    else:
        conts = intervals

    
    Lx = (xMaxWeb - xMinWeb)/size
    Ly = (yMaxWeb - yMinWeb)/size

    sh_total = []
    for cont in conts:
        
        lines = find_contours(raster[bandNumber-1,:,:],  mask=raster[-1,:,:] == 1, level = cont)
         
        newLines = []
        line = lines[0]
        for line in lines:
            newLine = LineString([ Point( xMinWeb + x[0] * Lx  , yMinWeb +  x[1] * Ly )  for x in line])
            newLines = newLines + [newLine]
        
        sh = gpd.GeoDataFrame({'geometry': newLines, 'label': (np.repeat(cont, len(newLines)))} )        
        
        sh_total = sh_total + [sh]
    
    sh = pd.concat(sh_total)
    sh.crs = 'EPSG:3857'
    sh = sh.to_crs('EPSG:' + str(epsg))

    return sh


def getValuesAlongLine(pathId, timestampId, line, token = None, epsg = 4326):
    pathId = sanitize.validUuid('pathId', pathId, True)
    timestampId = sanitize.validUuid('timestampId', timestampId, True)
    token = sanitize.validString('token', token, False)
    line = sanitize.validShapely('line', line, True)
    epsg = sanitize.validInt('epsg', epsg, True)

    if line.type != 'LineString':
        raise ValueError('line must be a shapely lineString')

    temp = gpd.GeoDataFrame({'geometry':[line]})
    temp.crs = 'EPSG:' + str(epsg)
    temp = temp.to_crs('EPSG:3857')
    line = temp['geometry'].values[0]
    line = list(line.coords)
        
    x_of_line = [p[0] for p in line]
    y_of_line = [p[1] for p in line]

    #the first action is to cacluate a bounding box for the raster we need to retrieve
    xMin = min(x_of_line)
    xMax = max(x_of_line)
    yMin = min(y_of_line)
    yMax = max(y_of_line)
    
    d = (xMax - xMin) * 0.1
    xMax = xMax + d
    xMin = xMin -d
    d = (yMax - yMin) * 0.1
    yMax = yMax + d
    yMin = yMin -d
    
    
    
    #now we retrieve the needed raster we use epsg = 4326 but we can use other coordinates as well
    extent = {'xMin': xMin, 'xMax':xMax, 'yMin':yMin, 'yMax':yMax}

    size = 1000
    r = getSampledRaster(pathId = pathId, timestampId = timestampId, extent = extent, width = size, height = size, epsg=3857, token = token)
    raster = r['raster']

    memfile =  MemoryFile()
    dataset = memfile.open( driver='GTiff', dtype='float32', height=size, width=size, count = raster.shape[0], crs= r['crs'], transform=r['transform'])
    dataset.write(raster)



    values = list(dataset.sample(line))

    memfile.close()
    return values


    
def getRaster(pathId, timestampId, extent, token = None, showProgress = True, epsg = 4326, reproject = False, threads = None, style = None):
    bounds = extent
    
    token = sanitize.validString('token', token, False)
    pathId = sanitize.validUuid('pathId', pathId, True)
    timestampId = sanitize.validUuid('timestampId', timestampId, True)
    bounds = sanitize.validBounds('bounds', bounds, True)
    showProgress = sanitize.validBool('showProgress', showProgress, True)
    reproject = sanitize.validBool('reproject', reproject, True)
    
    if type(threads)!= type(None):
        Warning('The parameter threads is deprecated')
    if type(style)!= type(None):
        Warning('The parameter style is deprecated')
        
    xMin = bounds['xMin']
    yMin = bounds['yMin']
    xMax = bounds['xMax']
    yMax = bounds['yMax']

    res = getActualExtent(xMin, xMax, yMin, yMax, 'EPSG:' + str(epsg))
    if res['status'] == '400':
        raise ValueError('Invalid epsg and extent combination')
        
    bounds = res['message']

    xMinWeb = bounds['xMin']
    yMinWeb = bounds['yMin']
    xMaxWeb = bounds['xMax']
    yMaxWeb = bounds['yMax']
    
    info = apiManager.get('/path/' + pathId, None, token)
    bands =  info['raster']['bands']
    as_jpg = info['raster']['asJpg']
    num_bands = len(bands)
    dtype = info['raster']['format']

    timestamps =  info['raster']['timestamps']
    all_timestamps = [item['id'] for item in timestamps]
    if not timestampId in all_timestamps:
        raise ValueError('given timestamp does not exist')
    
    t = next(item for item in timestamps if item["id"] == timestampId)

    if t['status'] != 'active':
        raise ValueError('timestamp is not active, please active timestamp before making queries to it')

    zoom = t['zoom']

    body = {}

    LEN = 2.003751e+07

    x_start = 2**zoom * (xMinWeb + LEN) / (2* LEN)
    x_end = 2**zoom * (xMaxWeb + LEN) / (2* LEN)
    y_end = 2**zoom * (LEN - yMinWeb) / (2* LEN)
    y_start = 2**zoom * (LEN - yMaxWeb) / (2* LEN)

    x1_osm = math.floor(x_start)
    x2_osm = math.floor(x_end)
    y1_osm = math.floor(y_start)
    y2_osm = math.floor(y_end)
    tarred = False
    
    #dit is voor testing beide moeten op false staan
    forceTar = False
    skipTar = False
    
    N=0
    w = 256
    if not skipTar and ((x2_osm - x1_osm) * (y2_osm - y1_osm) > 50 or forceTar):
        tarZoom = next(item for item in timestamps if item["id"] == timestampId)['tarZoom']
            
        if type(tarZoom) != type(None):
            N = zoom - tarZoom

            w = 256*2**N
            tarred = True
            zoom = tarZoom
            x_start = 2**zoom * (xMinWeb + LEN) / (2* LEN)
            x_end = 2**zoom * (xMaxWeb + LEN) / (2* LEN)
            y_end = 2**zoom * (LEN - yMinWeb) / (2* LEN)
            y_start = 2**zoom * (LEN - yMaxWeb) / (2* LEN)
        
            x1_osm = math.floor(x_start)
            x2_osm = math.floor(x_end)
            y1_osm = math.floor(y_start)
            y2_osm = math.floor(y_end)
            
    x_tiles = np.arange(x1_osm, x2_osm+1)
    y_tiles = np.arange(y1_osm, y2_osm +1)

    r_total = np.zeros((num_bands, w*(y2_osm - y1_osm + 1) ,w*(x2_osm - x1_osm + 1)), dtype = dtype)
    
    tiles = []            
    for tileY in y_tiles:
        for tileX in x_tiles:
            tiles = tiles + [(tileX, tileY)]
            
            
    def fetch(tileX, tileY):
        if tarred:

            cuts = cutOfTilesPerZoom[zoom]
            
            zones = [{'name': 'zone0', 'offset':0, 'start':cuts['start'][0] , 'end':cuts['end'][0]},  ]
            for i in [1,2,3]:
                zones = zones + [{'name':'zone' + str(i) + '_north','offset':i, 'start':cuts['start'][i] , 'end':cuts['start'][i-1] -1}, {'name':'zone' + str(i) + '_south','offset':i, 'start':cuts['end'][i-1]+1 , 'end':cuts['end'][i]}]

            for zone in zones:        
                if zone['start'] <= tileY  and tileY <= zone['end']:
                    offset = zone['offset']


            zoom_c = zoom - offset
            tileX_exact = tileX / 2**offset
            tileY_exact = tileY / 2**offset
            tileX_c = math.floor(tileX_exact)
            tileY_c = math.floor(tileY_exact)
            frac_x = int((tileX_exact - tileX_c)*w)
            frac_y = int(w*(tileY_exact - tileY_c))
            frac_w = int(w/2**offset)
            
            url = apiManager.baseUrl + '/path/' + pathId + '/raster/timestamp/' + timestampId + '/tarTile/' + str(zoom_c) + '/' + str(tileX_c) + '/' + str(tileY_c)

            if type(token) == type(None):
                r = requests.get(url)
            else:
                if not 'Bearer' in token:
                    r = requests.get(url, headers={"Authorization": 'Bearer ' + token})
                else:
                    r = requests.get(url, headers={"Authorization":  token})
            
        else:
            r = apiManager.get('/path/' + pathId + '/raster/timestamp/' + timestampId + '/tile/' + str(zoom) + '/' + str(tileX) + '/' + str(tileY), body, token, False)
            
        if r.status_code == 403:
                raise ValueError('insufficient access')
        elif r.status_code == 204:
                r = np.zeros((num_bands,w,w))
        elif r.status_code != 200:
                raise ValueError(r.text)
        else:
            if tarred:
                stream = BytesIO(r.content).read()
                
                r = constructImage(N, stream, as_jpg, num_bands, dtype)
                
                
                r = r[: ,frac_y:  frac_y + frac_w, frac_x: + frac_x + frac_w]
                r = np.transpose(r, [1,2,0])
                r = resize(r, (w,w), order = 0, preserve_range=True, mode = 'edge')
                r = np.transpose(r, [2,0,1])
                
            else:
                r = tifffile.imread(BytesIO(r.content))
                
        r = r.astype(dtype)            
        return r

            
            

    def iterate(fetch, i, tileX, tileY):
        return fetch(tileX, tileY)
        attempts = 5
        if i > 0:
            time.sleep(0.5)
        if i < attempts:
            try:
                return fetch(tileX, tileY)
            except:
                iterate(fetch, i+1, tileX, tileY)
        else:
            return fetch(tileX, tileY)
    I = 0

    tile = tiles[0]    
    for tile in tiles:
        tileX = tile[0]
        tileY = tile[1]
        x_index = tileX - x1_osm
        y_index = tileY - y1_osm
        
        r = iterate(fetch, 0, tileX, tileY)

        r_total[:,y_index*w:(y_index+1)*w,x_index*w:(x_index+1)*w] = r
        I = I + 1
        if showProgress:
            loadingBar(I, len(tiles))
        



    min_x_index = int(math.floor((x_start - x1_osm)*w))
    max_x_index = max(int(math.floor((x_end- x1_osm)*w + 1 )), min_x_index + 1 )
    min_y_index = int(math.floor((y_start - y1_osm)*w))
    max_y_index = max(int(math.floor((y_end- y1_osm)*w +1)), min_y_index + 1)

    r_total = r_total[:,min_y_index:max_y_index,min_x_index:max_x_index]

    mercatorExtent =  {'xMin' : xMinWeb, 'yMin': yMinWeb, 'xMax': xMaxWeb, 'yMax': yMaxWeb}
    if (not reproject) or epsg == 3857:
        trans = rasterio.transform.from_bounds(xMinWeb, yMinWeb, xMaxWeb, yMaxWeb, r_total.shape[2], r_total.shape[1])
    
        return {'raster': r_total, 'transform':trans, 'extent':mercatorExtent, 'epsg':3857}
    else:
        return reprojectRaster(r = r_total, sourceExtent = mercatorExtent, targetExtent = extent, targetWidth=r_total.shape[2], targetHeight=r_total.shape[1], sourceEpsg = 3857, targetEpsg= epsg, interpolation = 'nearest')


def analyse(pathId, timestampIds, geometry, returnType= 'all', approximate=True, token = None, epsg = 4326):
    token = sanitize.validString('token', token, False)
    pathId = sanitize.validUuid('pathId', pathId, True)    
    timestampIds = sanitize.validUuidArray('timestampIds', timestampIds, True)    
    approximate = sanitize.validBool(approximate, approximate, True)    
    geometry = sanitize.validShapely('geometry', geometry, True)
    returnType = sanitize.validString('returnType', returnType, True)

    temp = gpd.GeoDataFrame({'geometry':[geometry]})
    temp.crs = 'EPSG:' + str(epsg)
    temp = temp.to_crs('EPSG:4326')
    geometry = temp['geometry'].values[0]

    try:
        sh = gpd.GeoDataFrame({'geometry':[geometry]})
        geometry =sh.to_json(na='drop')
        geometry = json.loads(geometry)
        geometry = geometry['features'][0]['geometry']
    except:
        raise ValueError('geometry must be a shapely geometry')


    body = {'timestampIds':timestampIds, 'geometry':geometry, 'approximate':approximate, returnType:returnType}
    r = apiManager.get('/path/' + pathId + '/raster/timestamp/analyse', body, token)
    return r


def add(pathId, token, description= None, date ={'from': datetime.datetime.now(), 'to': datetime.datetime.now()}):
    

    token = sanitize.validString('token', token, True)
    pathId = sanitize.validUuid('pathId', pathId, True)  
    date = sanitize.validDateRange('date', date, True)    
    description = sanitize.validString('description', description, False)    

    body = { 'date': date, 'description': description}    
    r = apiManager.post('/path/' + pathId + '/raster/timestamp'  , body, token)
    return r
    
def edit(pathId, timestampId, token, date=None, description= None):

    
    
    token = sanitize.validString('token', token, True)
    pathId = sanitize.validUuid('pathId', pathId, True)  
    timestampId = sanitize.validUuid('timestampId', timestampId, True)  
    description = sanitize.validString('description', description, False)    
    date = sanitize.validDateRange('date', date, False)    

    body = {'date':date, 'description': description}    
    r = apiManager.patch('/path/' + pathId + '/raster/timestamp/' + timestampId  , body, token)
    return r

def getBounds(pathId, timestampId, token = None):
    token = sanitize.validString('token', token, False)
    pathId = sanitize.validUuid('pathId', pathId, True)  
    timestampId = sanitize.validUuid('timestampId', timestampId, True)  
    r = apiManager.get('/path/' + pathId + '/raster/timestamp/' + timestampId + '/bounds'  , None, token)
    r = {'id': 0, 'properties':{}, 'geometry':r}

    r  = gpd.GeoDataFrame.from_features([r])
    r = r.unary_union
    return r

def activate(pathId, timestampId, token):
    token = sanitize.validString('token', token, True)
    pathId = sanitize.validUuid('pathId', pathId, True)  
    timestampId = sanitize.validUuid('timestampId', timestampId, True)  
    r = apiManager.post('/path/' + pathId + '/raster/timestamp/' + timestampId + '/activate'  , None, token)
    return r

def deactivate(pathId, timestampId, token):
    token = sanitize.validString('token', token, True)
    pathId = sanitize.validUuid('pathId', pathId, True)  
    timestampId = sanitize.validUuid('timestampId', timestampId, True)  
    r = apiManager.post('/path/' + pathId + '/raster/timestamp/' + timestampId + '/deactivate'  , None, token)
    return r


def delete(pathId, timestampId, token):
    token = sanitize.validString('token', token, True)
    pathId = sanitize.validUuid('pathId', pathId, True)  
    timestampId = sanitize.validUuid('timestampId', timestampId, True)  

    r = apiManager.delete('/path/' + pathId + '/raster/timestamp/' + timestampId  , None, token)
    return r


def trash(pathId, timestampId, token):
    token = sanitize.validString('token', token, True)
    pathId = sanitize.validUuid('pathId', pathId, True)  
    timestampId = sanitize.validUuid('timestampId', timestampId, True)  
    body = {'trashed' : True}
    r = apiManager.put('/path/' + pathId + '/raster/timestamp/' + timestampId + '/trashed'  , body, token)
    return r

    
def recover(pathId, timestampId, token):
    token = sanitize.validString('token', token, True)
    pathId = sanitize.validUuid('pathId', pathId, True)  
    timestampId = sanitize.validUuid('timestampId', timestampId, True)  
    body = {'trashed' : False}
    r = apiManager.put('/path/' + pathId + '/raster/timestamp/' + timestampId + '/trashed'  , body, token)
    return r



from ellipsis import apiManager
from ellipsis import sanitize
from ellipsis.util.root import recurse
import geopandas as gpd
from shapely import geometry
import numpy as np

from ellipsis.util import chunks
from ellipsis.util import loadingBar
from ellipsis.util.root import stringToDate
from ellipsis.util.root import getActualExtent
from ellipsis.path import get as getInfo

import datetime

def add(pathId,  token, properties = None, description = None, date ={'from': datetime.datetime.now(), 'to': datetime.datetime.now()}):

    pathId = sanitize.validUuid('pathId', pathId, True) 
    token = sanitize.validString('token', token, True)
    date = sanitize.validDateRange('date', date, True)    
    properties = sanitize.validObject('properties', properties, False)
    description = sanitize.validString('description', description, False)

    body = {'properties':properties, 'date': date, 'description':description}
    r = apiManager.post('/path/' + pathId + '/vector/timestamp', body, token)
    return r

def edit(pathId, timestampId, token, description=None, date=None):


    
    
    pathId = sanitize.validUuid('pathId', pathId, True) 
    timestampId = sanitize.validUuid('timestampId', timestampId, True) 
    token = sanitize.validString('token', token, True)
    description = sanitize.validString('description', description, False)
    date = sanitize.validDateRange('date', date, False)    

    body = {'date':date,'description':description}
    r = apiManager.patch('/path/' + pathId + '/vector/timestamp/' + timestampId, body, token)
    return r
    

def trash(pathId, timestampId, token):
    pathId = sanitize.validUuid('pathId', pathId, True) 
    timestampId = sanitize.validUuid('timestampId', timestampId, True) 
    token = sanitize.validString('token', token, True)
    body = {'trashed':True}
    r = apiManager.put('/path/' + pathId + '/vector/timestamp/' + timestampId + '/trashed', body, token)
    return r


def recover(pathId, timestampId, token):
    pathId = sanitize.validUuid('pathId', pathId, True) 
    timestampId = sanitize.validUuid('timestampId', timestampId, True) 
    token = sanitize.validString('token', token, True)
    body = {'trashed':False}
    r = apiManager.put('/path/' + pathId + '/vector/timestamp/' + timestampId + '/trashed', body, token)
    return r

def delete(pathId, timestampId, token):
    pathId = sanitize.validUuid('pathId', pathId, True) 
    timestampId = sanitize.validUuid('timestampId', timestampId, True) 
    token = sanitize.validString('token', token, True)
    r = apiManager.delete('/path/' + pathId + '/vector/timestamp/' + timestampId , None, token)
    return r
    

def activate(pathId, timestampId, token):
    token = sanitize.validString('token', token, True)
    pathId = sanitize.validUuid('pathId', pathId, True)  
    timestampId = sanitize.validUuid('timestampId', timestampId, True)  
    r = apiManager.post('/path/' + pathId + '/vector/timestamp/' + timestampId + '/activate'  , None, token)
    return r

def deactivate(pathId, timestampId, token):
    token = sanitize.validString('token', token, True)
    pathId = sanitize.validUuid('pathId', pathId, True)  
    timestampId = sanitize.validUuid('timestampId', timestampId, True)  
    r = apiManager.post('/path/' + pathId + '/vector/timestamp/' + timestampId + '/deactivate'  , None, token)
    return r
    
def getBounds(pathId, timestampId, token = None):
    pathId = sanitize.validUuid('pathId', pathId, True) 
    timestampId = sanitize.validUuid('timestampId', timestampId, True) 
    token = sanitize.validString('token', token, False)
    r = apiManager.get('/path/' + pathId + '/vector/timestamp/' + timestampId + '/bounds' , None, token)

    r = {'id': 0, 'properties':{}, 'geometry':r}
    r  = gpd.GeoDataFrame.from_features([r])
    r = r.unary_union

    return r


def getChanges(pathId, timestampId, token = None, pageStart = None, listAll = False, actions = None):
    pathId = sanitize.validUuid('pathId', pathId, True) 
    timestampId = sanitize.validUuid('timestampId', timestampId, True) 
    token = sanitize.validString('token', token, False)
    listAll = sanitize.validBool('listAll', listAll, True)
    pageStart = sanitize.validObject('pageStart', pageStart, False) 
    actions = sanitize.validObject('actions', actions, False)
    body = {'pageStart':pageStart}    
    def f(body):
        r = apiManager.get('/path/' + pathId + '/vector/timestamp/' + timestampId + '/changelog' , body, token)
        return r
    r = recurse(f, body, listAll)
    r['result'] = [ {**x, 'date':stringToDate(x['date'])} for x in r['result'] ]
    return r



def getFeaturesByIds(pathId, timestampId, featureIds, token = None, showProgress = True, levelOfDetail = None):
    pathId = sanitize.validUuid('pathId', pathId, True) 
    timestampId = sanitize.validUuid('timestampId', timestampId, True) 
    token = sanitize.validString('token', token, False)
    featureIds = sanitize.validUuidArray('featureIds', featureIds, True)
    showProgress = sanitize.validBool('showProgress', showProgress, True)
    levelOfDetail = sanitize.validInt('levelOfDetail', levelOfDetail, False)
    
    id_chunks = chunks(featureIds, 10)

    r = {'size': 0 , 'result': [], 'nextPageStart' : None}
    ids = id_chunks[0]
    i=0
    for ids in id_chunks:
        body = {'geometryIds': ids, levelOfDetail:levelOfDetail}
        r_new = apiManager.get('/path/' + pathId + '/vector/timestamp/' + timestampId + '/featuresByIds' , body, token)
        
        r['result'] = r['result'] + r_new['result']['features']
        r['size'] = r['size'] + r_new['size']
        if len(id_chunks) >0 and showProgress:
            loadingBar(i*10 + len(ids),len(featureIds))
        i=i+1

    try:
        sh = gpd.GeoDataFrame.from_features(r['result']['features'])
    except:
        for i in range(len(r['result']['features'])):
            if 'crs' in r['result']['features'][i]['properties'].keys():
                del r['result']['features'][i]['properties']['crs']
        sh = gpd.GeoDataFrame.from_features(r['result']['features'])
        

    r['result'] = sh
    return r
    

def getFeaturesByExtent(pathId, timestampId, extent, propertyFilter = None, token = None, listAll = True, pageStart = None, epsg = 4326, coordinateBuffer = None, levelOfDetail = None):
    pathId = sanitize.validUuid('pathId', pathId, True) 
    timestampId = sanitize.validUuid('timestampId', timestampId, True) 
    token = sanitize.validString('token', token, False)
    extent = sanitize.validBounds('extent', extent, True)
    levelOfDetail = sanitize.validInt('levelOfDetail', levelOfDetail, False)
    propertyFilter = sanitize.validObject('propertyFilter', propertyFilter, False)
    listAll = sanitize.validBool('listAll', listAll, True)
    epsg = sanitize.validInt('epsg', epsg, True)
    pageStart = sanitize.validObject('pageStart', pageStart, False) 
    coordinateBuffer = sanitize.validFloat('coordinateBuffer', coordinateBuffer, False) 
    if str(type(coordinateBuffer)) == str(type(None)):
        info = getInfo(pathId, token)
        ts = [x for x in info['vector']['timestamps'] if x['id'] == timestampId]
        if len(ts) == 0:
            raise ValueError('Given timestampId does not exist')
        t = ts[0]
        zoom = t['zoom']
        coordinateBuffer = 0.5*360 / 2**zoom


    extent_new = {}
    extent_new['xMin'] = max(-180, extent['xMin'] - coordinateBuffer)
    extent_new['xMax'] = min(180, extent['xMax'] + coordinateBuffer)
    extent_new['yMin'] = max(-85, extent['yMin'] - coordinateBuffer)
    extent_new['yMax'] = min(85, extent['yMax'] + coordinateBuffer)



    res = getActualExtent(extent_new['xMin'], extent_new['xMax'], extent_new['yMin'], extent_new['yMax'], 'EPSG:' + str(epsg), 4326)
    if res['status'] == '400':
        raise ValueError('Invalid epsg and extent combination')

    extent_new = res['message']


    body = {'pageStart': pageStart, 'propertyFilter':propertyFilter, 'extent':extent_new, 'levelOfDetail':levelOfDetail}

    def f(body):
        return apiManager.get('/path/' + pathId + '/vector/timestamp/' + timestampId + '/featuresByExtent' , body, token)
        
    r = recurse(f, body, listAll, 'features')
    
    try:
        sh = gpd.GeoDataFrame.from_features(r['result']['features'])
    except:
        for i in range(len(r['result']['features'])):
            if 'crs' in r['result']['features'][i]['properties'].keys():
                del r['result']['features'][i]['properties']['crs']
        sh = gpd.GeoDataFrame.from_features(r['result']['features'])

        
    if sh.shape[0] ==0:
        r['result'] = sh
        return r
    bounds = sh.bounds
    px = (bounds['minx'] + bounds['maxx']) /2
    py = (bounds['miny'] + bounds['maxy']) /2
    sh = sh[ np.logical_and(np.logical_and( px >= extent['xMin'], px <= extent['xMax'] ), np.logical_and( py >= extent['yMin'], py <= extent['yMax'] )  )  ]
    sh.crs = 'EPSG:4326'
    sh = sh.to_crs('EPSG:' + str(epsg))
    r['result'] = sh
    
    return r


def listFeatures(pathId, timestampId, token = None, listAll = True, pageStart = None, levelOfDetail = None):
    pathId = sanitize.validUuid('pathId', pathId, True) 
    timestampId = sanitize.validUuid('timestampId', timestampId, True) 
    token = sanitize.validString('token', token, False)
    listAll = sanitize.validBool('listAll', listAll, True)
    pageStart = sanitize.validObject('pageStart', pageStart, False) 
    levelOfDetail = sanitize.validInt('levelOfDetail', levelOfDetail, False)

    body = {'pageStart': pageStart, 'levelOfDetail':levelOfDetail, 'pageSize':10000}


    def f(body):
        if type(body['pageStart']) == type(None):
            try:
                return  apiManager.get('/path/' + pathId + '/vector/timestamp/' + timestampId + '/compressedListFeatures' , body, token)
            except:
                return apiManager.get('/path/' + pathId + '/vector/timestamp/' + timestampId + '/listFeatures' , body, token)                
        else:
            return apiManager.get('/path/' + pathId + '/vector/timestamp/' + timestampId + '/listFeatures' , body, token)

    r = recurse(f, body, listAll, 'features')



    try:
        sh = gpd.GeoDataFrame.from_features(r['result']['features'])
    except:
        for i in range(len(r['result']['features'])):
            if 'crs' in r['result']['features'][i]['properties'].keys():
                del r['result']['features'][i]['properties']['crs']
        sh = gpd.GeoDataFrame.from_features(r['result']['features'])
    

    sh.crs = {'init': 'epsg:4326'}
    r['result'] = sh

    return r







from ellipsis import apiManager
from ellipsis import sanitize
from ellipsis.util.root import recurse
from ellipsis.util.root import chunks
from ellipsis.util.root import loadingBar
from ellipsis.util.root import stringToDate


import numpy as np
import pandas as pd

def get(pathId, timestampId, featureId, pageStart = None, dateTo = None, userId = None, seriesProperty = None, deleted = False, listAll= True,  token = None ):    
    pathId = sanitize.validUuid('pathId', pathId, True) 
    timestampId = sanitize.validUuid('timestampId', timestampId, True) 
    featureId = sanitize.validUuid('featureId', featureId, True) 
    pageStart = sanitize.validUuid('pageStart', pageStart, False)
    dateTo = sanitize.validDate('dateTo', dateTo, False)
    userId = sanitize.validUuid('userId', userId, False)
    seriesProperty = sanitize.validString('seriesProperty', seriesProperty, False)
    deleted = sanitize.validBool('deleted', deleted, True)
    listAll = sanitize.validBool('listAll', listAll, True)
    token = sanitize.validString('token', token, False)

    body = {'pageStart' : pageStart, 'dateTo' : dateTo, 'userId' : userId, 'seriesProperty': seriesProperty, 'deleted': deleted, 'listAll': listAll}
    
    def f(body):
        r = apiManager.get('/path/' + pathId + '/vector/timestamp/' + timestampId + '/feature/' + featureId + '/series/element', body, token)
        return r
    
    r = recurse(f, body, listAll)
    series = r['result']
    series = [ { 'id':k['id'], 'property': k['property'], 'value': k['value'], 'date': stringToDate(k['date']) } for k in series]
    series = pd.DataFrame(series)
    r['result'] = series

    return r
    

def info(pathId, timestampId, featureId, token = None):
    pathId = sanitize.validUuid('pathId', pathId, True) 
    timestampId = sanitize.validUuid('timestampId', timestampId, True) 
    featureId = sanitize.validUuid('featureId', featureId, True) 
    token = sanitize.validString('token', token, False)
    
    r = apiManager.get('/path/' + pathId + '/vector/timestamp/' + timestampId + '/feature/' + featureId + '/series/info', None, token)
    r['dateFrom'] = stringToDate(r['dateFrom'])
    r['dateTo'] = stringToDate(r['dateTo'])
    return r


def add(pathId, timestampId, seriesData, token, featureId=None, showProgress = True):
    pathId = sanitize.validUuid('pathId', pathId, True) 
    timestampId = sanitize.validUuid('timestampId', timestampId, True) 
    featureId = sanitize.validUuid('featureId', featureId, False)
    token = sanitize.validString('token', token, True)
    seriesData = sanitize.validDataframe('seriesData', seriesData, True)
    showProgress = sanitize.validBool('showProgress', showProgress, True)

    if str(type(featureId)) != str(type(None)):
        seriesData['featureId'] = featureId

    if 'date' in seriesData.columns:
        if 'datetime64' in str(seriesData['date'].dtypes):
            seriesData['date'] = seriesData['date'].dt.strftime('%Y-%m-%d %H:%M:%S')
            dates = list(seriesData['date'])
            del seriesData['date']
        else:
           raise  ValueError('date column must be of type datetime')
    else:
           raise  ValueError('seriesData must have a column datetime of type datetime')

    if not 'featureId' in seriesData.columns:
        raise ValueError(
            "You either need to supply a featureId in the function parameters or supply a featureId column in seriesData")

    featureIds = seriesData['featureId']
    del seriesData['featureId']

    for c in seriesData.columns:
            seriesData[c] = seriesData[c].astype(float)
        

    values = []
    for c in seriesData.columns:
        values = values +  [{'property': c, 'featureId': x[0], 'value': x[1], 'date': x[2]} for x in zip(featureIds, seriesData[c].values, dates) if not np.isnan(x[1]) ]


    chunks_values = chunks(values)
    N = 0
    r_total = []
    for values_sub in chunks_values:
        body = { "values":values_sub}
        r = apiManager.post("/path/" + pathId + "/vector/timestamp/" + timestampId  + '/feature/series/element', body, token)
        
        r_total = r_total + r
        if len(chunks_values) >1 and showProgress:
            loadingBar(N*3000 + len(values_sub), len(values))
        N = N+1
    return r_total




def trash(pathId, timestampId, featureId, seriesIds, token, showProgress = True):
    pathId = sanitize.validUuid('pathId', pathId, True) 
    timestampId = sanitize.validUuid('timestampId', timestampId, True) 
    featureId = sanitize.validUuid('featureId', featureId, True) 
    token = sanitize.validString('token', token, True)
    seriesIds = sanitize.validUuidArray('seriesIds', seriesIds, True)
    showProgress = sanitize.validBool('showProgress', showProgress, True)

    chunks_values = chunks(seriesIds)
    N = 0
    for seriesIds_sub in chunks_values:
        body = { "seriesIds":seriesIds_sub, "trashed": True}
        r = apiManager.put("/path/" + pathId + "/vector/timestamp/" + timestampId  + '/feature/' + featureId + '/series/element/trashed', body, token)
        if showProgress:
            loadingBar(N*3000 + len(seriesIds_sub), len(seriesIds))
        N = N+1
    return r

def recover(pathId, timestampId, featureId, seriesIds, token, showProgress = True):
    pathId = sanitize.validUuid('pathId', pathId, True) 
    timestampId = sanitize.validUuid('timestampId', timestampId, True) 
    featureId = sanitize.validUuid('featureId', featureId, True) 
    token = sanitize.validString('token', token, True)
    seriesIds = sanitize.validUuidArray('seriesIds', seriesIds, True)
    showProgress = sanitize.validBool('showProgress', showProgress, True)

    chunks_values = chunks(seriesIds)
    N = 0

    for seriesIds_sub in chunks_values:
        body = { "seriesIds":seriesIds_sub, "trashed": False}
        r = apiManager.put("/path/" + pathId + "/vector/timestamp/" + timestampId  + '/feature/' + featureId + '/series/element/trashed', body, token)
        if showProgress:           
            loadingBar(N*3000 + len(seriesIds_sub), len(seriesIds))
        N = N+1
    return r


def changelog(pathId, timestampId, featureId, listAll = False, actions = None, userId = None, pageStart = None, token = None):
    pathId = sanitize.validUuid('pathId', pathId, True) 
    timestampId = sanitize.validUuid('timestampId', timestampId, True) 
    featureId = sanitize.validUuid('featureId', featureId, True) 
    listAll = sanitize.validBool('listAll', listAll, True) 
    actions = sanitize.validStringArray('actions', actions, False) 
    userId = sanitize.validUuid('userId', userId, False) 
    pageStart = sanitize.validUuid('pageStart', pageStart, False) 
    token = sanitize.validString('token', token, False)

    body = {'userId': userId, 'actions':actions, 'pageStart':pageStart}

    def f(body):
        r = apiManager.get('/path/' + pathId + '/vector/timestamp/' + timestampId + '/feature/' + featureId + '/series/changelog', body, token)
        return r
        
    r = recurse(f, body, listAll)

    r['result'] = [{**x, 'date':stringToDate(x['date'])} for x in r['result']]
    return r

    
    



from ellipsis import apiManager
from ellipsis import sanitize
from ellipsis.util.root import recurse
from ellipsis.util.root import stringToDate
from ellipsis.path.folder.root import listFolder

def search(pathTypes = ['raster', 'vector', 'file', 'folder'] ,root=None, text=None, active=None, userId=None, pageStart=None, hashtag=None, extent=None, resolution=None, date=None, listAll= False, token=None):

    token = sanitize.validString('token', token, False)
    pathTypes = sanitize.validStringArray('pathTypes', pathTypes, True)
    root = sanitize.validStringArray('root', root, False)
    text = sanitize.validString('text', text, False)
    userId = sanitize.validUuid('userId', userId, False)
    pageStart = sanitize.validUuid('pageStart', pageStart, False)
    hashtag = sanitize.validString('hashtag', hashtag, False)



    listAll = sanitize.validBool('listAll', listAll, True)

    active = sanitize.validBool('active', active, False)

    extent = sanitize.validBounds('extent', extent, False)
    
    date = sanitize.validDateRange('date', date, False)


    resolution = sanitize.validResolution('resolution', resolution, False)

    body = {
        'type': pathTypes,
        'root': root,
        'text': text,
        'userId': userId,
        'active': active,
        'pagestart': pageStart,
        'hashtag': hashtag,
        'extent': extent,
        'resolution': resolution,
        'date':date
    }
    def f(body):
        return apiManager.get('/path', body, token)
    
    r = recurse(f, body, listAll)
    return r



def get(pathId, token=None):
    pathId = sanitize.validUuid('pathId', pathId, True)
    token = sanitize.validString('token', token, False)
    r = apiManager.get(f"/path/{pathId}", None, token)
    r = convertPath(r)
    return r

def setDomains(pathId, token, domains):
    pathId = sanitize.validUuid('pathId', pathId, True)
    token = sanitize.validString('token', token, True)
    domains = sanitize.validStringArray('domains', domains, True)

    r = apiManager.get(f"/path/{pathId}/domain", None, token)
    return r


####depricated
def listPath(pathId, pathType, pageStart=None, listAll = True, token=None):
    pathId = sanitize.validUuid('pathId', pathId, True)
    token = sanitize.validString('token', token, False)
    pageStart = sanitize.validUuid('pageStart', pageStart, False)
    listAll = sanitize.validBool('listAll', listAll, False)

    if type(pathType) != type('x') or (pathType != 'folder' and pathType !='layer' ) :
        raise ValueError("pathType must be one of 'layer' or 'folder'")
        
    isFolder = False
    if pathType == 'folder':
        isFolder = True        


    body  = {'pageStart': pageStart, 'isFolder':isFolder}
    
    def f(body):
        
        return apiManager.get(f'/path/{pathId}/list', body, token)

    r = recurse(f, body, listAll)
    return r


def editMetadata(pathId, token, description=None, attribution=None, licenseString =None, properties=None):
    pathId = sanitize.validUuid('pathId', pathId, True)
    token = sanitize.validString('token', token, True)
    attribution = sanitize.validString('attribution', attribution, False)
    description = sanitize.validString('description', description, False)
    properties = sanitize.validObject('properties', properties, False)
    licenseString = sanitize.validString('licenseString', licenseString, False)
    return apiManager.patch(f'/path/{pathId}/metadata', {
        'description': description,
        'attribution': attribution,
        'properties': properties,
        'license': licenseString
    }, token)

#####depricated
def add( pathType, name, token, parentId = None, publicAccess =None, metadata=None):
    pathType = sanitize.validString('pathType', pathType, True)
    name = sanitize.validString('name', name, True)
    token = sanitize.validString('token', token, True)
    parentId = sanitize.validUuid('parentId', parentId, False)
    metadata = sanitize.validObject('metadata', metadata, False)
    publicAccess = sanitize.validObject('publicAccess', publicAccess, False)

    body = {'name': name, 'parentId':parentId, 'type': pathType, 'publicAccess':publicAccess, 'metadata':metadata }

    return apiManager.post('/path', body, token)

def rename(pathId, name, token):
    pathId = sanitize.validUuid('pathId', pathId, True)
    token = sanitize.validString('token', token, False)
    name = sanitize.validString('name', name, False)

    return apiManager.put(f'/path/{pathId}/name', {
        'name': name
    }, token)


def move(pathIds, parentId, token):
    token = sanitize.validString('token', token, False)
    pathIds = sanitize.validUuidArray('pathIds', pathIds, True)
    parentId = sanitize.validUuid('parentId', parentId, False)

    return apiManager.put('/path/parentId', {
        'pathIds': pathIds,
        'parentId': parentId,
    }, token)


def trash(pathId, token):
    pathId = sanitize.validUuid('pathId', pathId, True)
    token = sanitize.validString('token', token, False)

    return apiManager.put(f'/path/{pathId}/trashed', {
        'trashed': True
    }, token)


def recover(pathId, token):
    pathId = sanitize.validUuid('pathId', pathId, True)
    token = sanitize.validString('token', token, False)

    return apiManager.put(f'/path/{pathId}/trashed', {
        'trashed': False
    }, token)


def delete(pathId, token, recursive = False):
    pathId = sanitize.validUuid('pathId', pathId, True)
    token = sanitize.validString('token', token, False)
    recursive = sanitize.validBool('recursive', recursive, True)
    
    if recursive:
        info = get(pathId, token)
        if info['type'] == 'folder':
            folders = listFolder(pathId = pathId, includeTrashed=True, pathTypes=['folder'], token=token)['result']
            for f in folders:
                delete(f['id'], token, True)
            maps = listFolder(pathId=pathId, pathTypes=['raster','vector','file', 'pointCloud'], includeTrashed=True, token=token)['result']
            for m in maps:
                delete(m['id'], token, True)
        apiManager.delete(f'/path/{pathId}', None, token)            
    
    else:
        return apiManager.delete(f'/path/{pathId}', None, token)

def editPublicAccess(pathId, token, access = None, hidden=None, recursive = False):
    pathId = sanitize.validUuid('pathId', pathId, True)
    token = sanitize.validString('token', token, False)

    access = sanitize.validObject('access', access, False)
    hidden = sanitize.validBool('hidden', hidden, False)
    recursive = sanitize.validBool('recursive', recursive, True)

    if type(hidden) == type(None) and type(access) == type(None):
        raise ValueError('hidden and access cannot both be None')

    body = access
    body['hidden'] =  hidden



    if recursive:
        info = get(pathId, token)
        if info['type'] == 'folder':
            folders = listFolder(pathId=pathId, includeTrashed=False, pathTypes=['folder'], token=token)['result']
            for f in folders:
                editPublicAccess(pathId = f['id'], token = token, recursive = True, hidden = hidden, access = access)
            maps =  listFolder(pathId=pathId, pathTypes=['raster', 'vector', 'file', 'pointCloud'], includeTrashed=False, token=token)['result']
            for m in maps:
                editPublicAccess(pathId = m['id'], token = token, access=access, hidden = hidden, recursive = True)
        return apiManager.patch('/path/' + pathId + '/publicAccess', body, token)


    else:
        return apiManager.patch('/path/' + pathId + '/publicAccess', body, token)
    
def favorite(pathId, token):
    pathId = sanitize.validUuid('pathId', pathId, True)
    token = sanitize.validString('token', token, False)
    body = {'favorite':True}
    return apiManager.put(f'/path/{pathId}/favorite', body, token)
    
    
def unfavorite(pathId, token):
    pathId = sanitize.validUuid('pathId', pathId, True)
    token = sanitize.validString('token', token, False)
    body = {'favorite':False}
    return apiManager.put(f'/path/{pathId}/favorite', body, token)
    
##helper function
    
def convertPath(path):
    if path['type'] == 'raster':
        path['raster']['timestamps'] = [ {**x, 'date' : {'from':stringToDate(x['date']['from']), 'to' : stringToDate(x['date']['to']) }} for x in path['raster']['timestamps'] ]
    if path['type'] == 'vector':
        path['vector']['timestamps'] = [ {**x, 'date' : {'from':stringToDate(x['date']['from']), 'to' : stringToDate(x['date']['to']) }} for x in path['vector']['timestamps'] ]
        
    return(path)




from ellipsis import apiManager
from ellipsis import sanitize
from ellipsis.util import chunks
from ellipsis.util import loadingBar
from ellipsis.util.root import stringToDate
from ellipsis.util.root import recurse

import numpy as np
import json
import geopandas as gpd


def manageLevels(levelOfDetail1, levelOfDetail2, levelOfDetail3, levelOfDetail4, levelOfDetail5, features):
    levelOfDetail1 = sanitize.validGeoSeries('levelOfDetail1', levelOfDetail1, False)
    levelOfDetail2 = sanitize.validGeoSeries('levelOfDetail2', levelOfDetail2, False)
    levelOfDetail3 = sanitize.validGeoSeries('levelOfDetail3', levelOfDetail3, False)
    levelOfDetail4 = sanitize.validGeoSeries('levelOfDetail4', levelOfDetail4, False)
    levelOfDetail5 = sanitize.validGeoSeries('levelOfDetail5', levelOfDetail5, False)

    if type(levelOfDetail1) != type(None) and len(levelOfDetail1) != features.shape[0]:
        raise ValueError('levelsOfDetail1 must have same length as number of rows in features')
    if type(levelOfDetail2) != type(None) and len(levelOfDetail2) != features.shape[0]:
        raise ValueError('levelsOfDetail2 must have same length as number of rows in features')
    if type(levelOfDetail3) != type(None) and len(levelOfDetail3) != features.shape[0]:
        raise ValueError('levelsOfDetail3 must have same length as number of rows in features')
    if type(levelOfDetail4) != type(None) and len(levelOfDetail4) != features.shape[0]:
        raise ValueError('levelsOfDetail4 must have same length as number of rows in features')
    if type(levelOfDetail5) != type(None) and len(levelOfDetail5) != features.shape[0]:
        raise ValueError('levelsOfDetail5 must have same length as number of rows in features')

    if type(levelOfDetail2) != type(None) and type(levelOfDetail1) == type(None):
        raise ValueError('If levelOfDetail2 is defined so should levelOfDetail1')
    if type(levelOfDetail3) != type(None) and type(levelOfDetail2) == type(None):
        raise ValueError('If levelOfDetail3 is defined so should levelOfDetail2')
    if type(levelOfDetail4) != type(None) and type(levelOfDetail3) == type(None):
        raise ValueError('If levelOfDetail4 is defined so should levelOfDetail3')
    if type(levelOfDetail5) != type(None) and type(levelOfDetail4) == type(None):
        raise ValueError('If levelOfDetail5 is defined so should levelOfDetail4')


    if type(levelOfDetail1) != type(None):
        levelOfDetail1 = levelOfDetail1.to_json()
        levelOfDetail1 = json.loads(levelOfDetail1)
        levelOfDetail1 = np.array( [ x['geometry'] for x in levelOfDetail1['features'] ])
    if type(levelOfDetail2) != type(None):
        levelOfDetail2 = levelOfDetail2.to_json()
        levelOfDetail2 = json.loads(levelOfDetail2)
        levelOfDetail2 = np.array( [ x['geometry'] for x in levelOfDetail2['features'] ])
    if type(levelOfDetail3) != type(None):
        levelOfDetail3 = levelOfDetail3.to_json()
        levelOfDetail3 = json.loads(levelOfDetail3)
        levelOfDetail3 = np.array( [ x['geometry'] for x in levelOfDetail3['features'] ])
    if type(levelOfDetail4) != type(None):
        levelOfDetail4 = levelOfDetail4.to_json()
        levelOfDetail4 = json.loads(levelOfDetail4)
        levelOfDetail4 = np.array( [ x['geometry'] for x in levelOfDetail4['features'] ])
    if type(levelOfDetail5) != type(None):
        levelOfDetail5 = levelOfDetail5.to_json()
        levelOfDetail5 = json.loads(levelOfDetail5)
        levelOfDetail5 = np.array( [ x['geometry'] for x in levelOfDetail5['features'] ])
        
    return levelOfDetail1, levelOfDetail2, levelOfDetail3, levelOfDetail4, levelOfDetail5

def zipLevels(levelOfDetail1, levelOfDetail2, levelOfDetail3, levelOfDetail4, levelOfDetail5, indices_sub):
    if type(levelOfDetail5) != type(None):         
        levels = list(zip(levelOfDetail1[indices_sub], levelOfDetail2[indices_sub], levelOfDetail3[indices_sub], levelOfDetail4[indices_sub], levelOfDetail5[indices_sub]))
    elif type(levelOfDetail4) != type(None):         
        levels = list(zip(levelOfDetail1[indices_sub], levelOfDetail2[indices_sub], levelOfDetail3[indices_sub], levelOfDetail4[indices_sub]))
    elif type(levelOfDetail3) != type(None):         
        levels = list(zip(levelOfDetail1[indices_sub], levelOfDetail2[indices_sub], levelOfDetail3[indices_sub]))
    elif type(levelOfDetail2) != type(None):         
        levels = list(zip(levelOfDetail1[indices_sub], levelOfDetail2[indices_sub]))                      
    elif type(levelOfDetail1) != type(None):         
        levels = list(zip(levelOfDetail1[indices_sub]))
    else:
        levels = None
    return levels
    
def add(pathId, timestampId, features, token, showProgress = True, levelOfDetail1 = None, levelOfDetail2 = None, levelOfDetail3 = None, levelOfDetail4 = None, levelOfDetail5 = None, cores = 1):
    pathId = sanitize.validUuid('pathId', pathId, True)
    timestampId = sanitize.validUuid('timestampId', timestampId, True)
    token = sanitize.validString('token', token, True)
    features = sanitize.validGeopandas('features', features, True, cores = cores)
    showProgress = sanitize.validBool('showProgress', showProgress, True)
    cores = sanitize.validInt('cores', cores, True)
    
    levelOfDetail1, levelOfDetail2, levelOfDetail3, levelOfDetail4, levelOfDetail5 = manageLevels(levelOfDetail1, levelOfDetail2, levelOfDetail3, levelOfDetail4, levelOfDetail5, features)
    features_json = features.to_json(na='drop')
    features_json = json.loads(features_json)
    features_json = np.array(features_json['features'])
    
    #check if first time
    firstTime = apiManager.get('/path/' + pathId, None, token)
    if not 'vector' in firstTime.keys():
        raise ValueError('Can only add features if path is of type vector')
    firstTime = len(firstTime['vector']['properties']) ==0
    
    if firstTime:
        print('no properties known for this timestamp adding them automatically')
        columns = features.columns
        columns = [c for c in columns if c != 'geometry']
        for c in columns:
            if 'int' in str(features.dtypes[c]) or 'Int' in str(features.dtypes[c]):
                propertyType = 'integer'
            elif 'float' in str(features.dtypes[c]) or 'Float' in str(features.dtypes[c]):
                propertyType = 'float'
            elif 'bool' in str(features.dtypes[c]):
                propertyType = 'boolean'
            elif 'datetime' in str(features.dtypes[c]):
                propertyType = 'datetime'
            else:
                propertyType = 'string'


            ###date
            body = {'name': c , 'type': propertyType , 'required': False, 'private': False}
            apiManager.post('/path/' + pathId + '/vector/property', body, token)
    indices = chunks(np.arange(features.shape[0]))


    addedIds = []
    for i in np.arange(len(indices)):
        indices_sub = indices[i]
        features_sub = features_json[indices_sub]

        levels = zipLevels(levelOfDetail1, levelOfDetail2, levelOfDetail3, levelOfDetail4, levelOfDetail5, indices_sub)
                      
        if type(levels) != type(None):
            featuresBody = [{'feature': features_sub[i], 'levelsOfDetail': levels[i] } for i in np.arange(len(indices_sub))]
        else:
            featuresBody = [{'feature': features_sub[i] } for i in np.arange(len(indices_sub))]
        body = {"features":featuresBody}
        r = apiManager.post('/path/' + pathId + '/vector/timestamp/' + timestampId + '/feature', body, token)
        addedIds = addedIds + r
        if showProgress:
            loadingBar(i*3000 + len(indices_sub),features.shape[0])
        i = i+1
        
    return(addedIds)


    
def edit(pathId, timestampId, featureIds, token, features = None, showProgress = True, levelOfDetail1 = None, levelOfDetail2 = None, levelOfDetail3 = None, levelOfDetail4 = None, levelOfDetail5 = None, cores = 1):
    pathId = sanitize.validUuid('pathId', pathId, True) 
    timestampId = sanitize.validUuid('timestampId', timestampId, True) 
    token = sanitize.validString('token', token, True)
    features = sanitize.validGeopandas('features', features, False, cores=cores)
    featureIds = sanitize.validUuidArray('featureIds', featureIds, True)
    showProgress = sanitize.validBool('showProgress', showProgress, True)
    cores = sanitize.validInt('cores', cores, True)
    
    levelOfDetail1, levelOfDetail2, levelOfDetail3, levelOfDetail4, levelOfDetail5 = manageLevels(levelOfDetail1, levelOfDetail2, levelOfDetail3, levelOfDetail4, levelOfDetail5, features)


    if type(features) != type(None) and features.shape[0] != len(featureIds):
        raise ValueError('featureIds must be of same length as the features geopandas dataframe')
        

    indices = chunks(np.arange(len(featureIds)),1000)
    i=0
    for i in np.arange(len(indices)):
        indices_sub = indices[i]
        featureIds_sub = list( np.array(featureIds)[indices_sub])

        if type(features) != type(None):
            features_sub = features.iloc[indices_sub]        
            features_sub =features_sub.to_json(na='drop')
            features_sub = json.loads(features_sub)

        levels = zipLevels(levelOfDetail1, levelOfDetail2, levelOfDetail3, levelOfDetail4, levelOfDetail5, indices_sub)

        if type(levels) == type(None):
                changes = [{'featureId':x[0] , 'newProperties':x[1]['properties'], 'newGeometry':x[1]['geometry']} for x in zip(featureIds_sub, features_sub['features'])]
        else:
                changes = [{'featureId':x[0], 'levelsOfDetail': levels , 'newProperties':x[1]['properties'], 'newGeometry':x[1]['geometry']} for x in zip(featureIds_sub, features_sub['features'])]
            
        body = {'changes':changes}
        r = apiManager.patch('/path/' + pathId + '/vector/timestamp/' + timestampId + '/feature', body, token)

        if len(indices) > 1 and showProgress:
            loadingBar(i*1000 + len(indices_sub),len(featureIds))

    return r


def trash(pathId, timestampId, featureIds, token, showProgress = True):
    pathId = sanitize.validUuid('pathId', pathId, True) 
    timestampId = sanitize.validUuid('timestampId', timestampId, True) 
    token = sanitize.validString('token', token, True)
    featureIds = sanitize.validUuidArray('featureIds', featureIds, True)
    showProgress = sanitize.validBool('showProgress', showProgress, True)

    indices = chunks(np.arange(len(featureIds)),1000)
    i=0
    for i in np.arange(len(indices)):
        indices_sub = indices[i]
        featureIds_sub = list(np.array(featureIds)[indices_sub])

        body = {'featureIds': featureIds_sub, 'trashed': True}
        r = apiManager.put('/path/' + pathId + '/vector/timestamp/' + timestampId + '/feature/trashed', body, token)

        if len(indices) > 1 and showProgress:
            loadingBar(i*1000 + len(indices_sub),len(featureIds))

    return r


def recover(pathId, timestampId, featureIds, token, showProgress = True):
    pathId = sanitize.validUuid('pathId', pathId, True) 
    timestampId = sanitize.validUuid('timestampId', timestampId, True) 
    token = sanitize.validString('token', token, True)
    featureIds = sanitize.validUuidArray('featureIds', featureIds, True)
    showProgress = sanitize.validBool('showProgress', showProgress, True)

    indices = chunks(np.arange(len(featureIds)),1000)
    i=0
    for i in np.arange(len(indices)):
        indices_sub = indices[i]
        featureIds_sub = list(np.array(featureIds)[indices_sub])

        body = {'featureIds': featureIds_sub, 'trashed': False}
        r = apiManager.put('/path/' + pathId + '/vector/timestamp/' + timestampId + '/feature/trashed', body, token)

        if len(indices) > 1 and showProgress:
            loadingBar(i*1000 + len(indices_sub),len(featureIds))

    return r



def versions(pathId, timestampId, featureId, token = None, pageStart = None, listAll = True):
    pathId = sanitize.validUuid('pathId', pathId, True) 
    timestampId = sanitize.validUuid('timestampId', timestampId, True) 
    token = sanitize.validString('token', token, False)
    featureId = sanitize.validUuid('featureId', featureId, True)
    pageStart = sanitize.validUuid('pageStart', pageStart, False)
    listAll = sanitize.validBool('listAll', listAll, True)
    
    body = {'returnType':'all'}
    def f(body):
        return apiManager.get('/path/' + pathId + '/vector/timestamp/' + timestampId + '/feature/' + featureId + '/version', body, token)

    r = recurse(f, body, listAll)

    features = [ x['feature'] for x in r['result']]
    dates = [stringToDate(x['date']) for x in r['result'] ]
    usernames = [x['user']['username'] for x in r['result'] ]
    userIds = [x['user']['id'] for x in r['result'] ]

    sh = gpd.GeoDataFrame.from_features(features)     
    sh['username'] = usernames
    sh['userId'] = userIds
    sh['dates'] = dates
    
    sh.crs = {'init': 'epsg:4326'}
    r['result'] = sh
    return(r)





from ellipsis import apiManager
from ellipsis import sanitize

def add( name, token, parentId = None, publicAccess =None, metadata=None):
    name = sanitize.validString('name', name, True)
    token = sanitize.validString('token', token, True)
    parentId = sanitize.validUuid('parentId', parentId, False)
    metadata = sanitize.validObject('metadata', metadata, False)
    publicAccess = sanitize.validObject('publicAccess', publicAccess, False)

    body = {'name': name, 'parentId':parentId, 'publicAccess':publicAccess, 'metadata':metadata }

    return apiManager.post('/path/vector', body, token)


def editFilter(pathId, propertyFilter, token):
    pathId = sanitize.validUuid('pathId', pathId, True) 
    token = sanitize.validString('token', token, True)
    propertyFilter = sanitize.validObject('propertyFilter', propertyFilter, True)
    
    body = {'filter': propertyFilter}
    r = apiManager.post('/path/' + pathId + '/vector/filter' , body, token)
    return r






from ellipsis import apiManager
from ellipsis import sanitize
from ellipsis.util.root import recurse

def listViews(token):

    token = sanitize.validString('token', token, True)



    r = apiManager.get('/view', {}, token)

    return r

def get(viewId, token=None):
    token = sanitize.validString('token', token, False)
    viewId = sanitize.validString('viewId', viewId, True)

    r = apiManager.get('/view/' + viewId, {}, token)
        
    return r



def add(pathId, layers, name = None, persistent=False, dems=[], token = None):
    token = sanitize.validString('token', token, False)
    pathId = sanitize.validString('pathId', pathId, True)
    layers = sanitize.validObject('layers', layers, True)
    dems = sanitize.validObject('dems', dems, True)
    name = sanitize.validString('pathId', name, False)
    persistent = sanitize.validBool('persistent', persistent, True)

    r = apiManager.post('/view', {'name':name, 'layers':layers, 'dems':dems, 'pathId':pathId, 'persistent':persistent}, token)
        
    return r


def edit(viewId, token, layers=None, name=None, dems=None):
    viewId = sanitize.validString('viewId', viewId, True)
    token = sanitize.validString('token', token, False)
    layers = sanitize.validObject('layers', layers, False)
    dems = sanitize.validObject('dems', dems, False)
    name = sanitize.validString('pathId', name, False)

    r = apiManager.patch('/view/' + viewId, {'layers':layers, 'dems':dems, 'name':name}, token)
        
    return r



def delete(viewId, token):
    viewId = sanitize.validString('viewId', viewId, True)
    token = sanitize.validString('token', token, False)

    r = apiManager.delete('/view/' + viewId, {}, token)

        
    return r




from ellipsis import apiManager
from ellipsis import sanitize

def add(pathId, name, method, parameters, token, default=True):
    token = sanitize.validString('token', token, True)
    pathId = sanitize.validUuid('pathId', pathId, True)    
    method = sanitize.validString('method', method, True)    
    name = sanitize.validString('name', name, True)    
    default = sanitize.validBool('default', default, True)    
    parameters = sanitize.validObject('parameters', parameters, True)

    body = {'name':name, 'method': method, 'parameters':parameters, 'default':default}
    r = apiManager.post('/path/' + pathId + '/raster/style', body, token)
    return r

    
def delete(pathId, styleId, token):
    token = sanitize.validString('token', token, True)
    pathId = sanitize.validUuid('pathId', pathId, True)    
    styleId = sanitize.validUuid('styleId', styleId, True)    

    r = apiManager.delete('/path/' + pathId + '/raster/style/' + styleId, None, token)
    return r


def edit(pathId, styleId, method, parameters, token, default = None):
    token = sanitize.validString('token', token, True)
    pathId = sanitize.validUuid('pathId', pathId, True)    
    method = sanitize.validString('method', method, True)    
    parameters = sanitize.validObject('parameters', parameters, True)
    default = sanitize.validBool('default', default, False)    

    body = {'method': method, 'parameters':parameters, 'default':default}
    r = apiManager.patch('/path/' + pathId + '/raster/style/' + styleId, body, token)
    return r




import geopandas as gpd
from shapely.geometry import Polygon
from geopy.distance import geodesic
import numpy as np
import pandas as pd
import math
import sys
import open3d as o3d

from datetime import datetime
from shapely import geometry
import rasterio
import matplotlib.pyplot as plt
from PIL import Image
import pygltflib
from skimage.transform import resize
import os
import random
import subprocess
from concurrent.futures import ProcessPoolExecutor as Pool
import multiprocessing
import warnings
import uuid
import shapely
from shapely.ops import cascaded_union


from ellipsis import sanitize
from rasterio.warp import reproject as warp, Resampling, calculate_default_transform

warnings.simplefilter("ignore")

def plotPointCloud(df, method = 'cloud', width = 800, height = 600, scale = 0.003):
    if type(method) != type('x'):
        raise ValueError('method must be one of cloud, voxel or mesh')
    if not method in ['cloud', 'mesh', 'voxel']:
        raise ValueError('method must be one of cloud, voxel or mesh')
    if type(df) != type(pd.DataFrame()):
        raise ValueError('df must be a dataframe with columns x,y,z, red, green,blue all of type float')


    maxx = np.max(df['x'].values)
    maxy = np.max(df['y'].values)
    maxz = np.max(df['z'].values)
    minx = np.min(df['x'].values)
    miny = np.min(df['y'].values)
    minz = np.min(df['z'].values)


    M = max(maxx - minx,maxy-miny, maxz-minz)


    points = [ [( p[0] - minx  )/M, (p[1]-miny)/M, (p[2] - minz)/M] for p in zip( df['x'].values, df['y'].values, df['z'].values)]
    colors = [  [p[0], p[1], p[2]]  for p in   zip( df['red'].values/ 255, df['green'].values/ 255, df['blue'].values/255)]

    # Initialize a point cloud object
    pcd = o3d.geometry.PointCloud()
    pcd.points = o3d.utility.Vector3dVector(points)
    pcd.colors = o3d.utility.Vector3dVector(colors)

    vis = o3d.visualization.Visualizer()

    vis.create_window(window_name='Point cloud', width=width, height=height)

    if method == 'cloud':
        # Add the pointcloud to the visualizer
        vis.add_geometry(pcd)
    elif method == 'voxel':
        # Create a voxel grid from the point cloud with a voxel_size of 0.01
        voxel_grid=o3d.geometry.VoxelGrid.create_from_point_cloud(pcd,voxel_size=scale)

        # Add the voxel grid to the visualizer
        vis.add_geometry(voxel_grid)
    elif method == 'mesh':
        mesh = o3d.geometry.TriangleMesh.create_from_point_cloud_alpha_shape(pcd, scale)
        mesh.compute_vertex_normals()
        vis.add_geometry(mesh)


    # We run the visualizater
    vis.run()
    vis.destroy_window()








def parseGlb(stream):
    gltf = pygltflib.GLTF2.load_from_bytes(stream)
    binary_blob = gltf.binary_blob()
    translation = gltf.nodes[0].translation

    points_accessor = gltf.accessors[gltf.meshes[0].primitives[0].attributes.POSITION]
    points_buffer_view = gltf.bufferViews[points_accessor.bufferView]
    points = np.frombuffer(
        binary_blob[
        points_buffer_view.byteOffset
        + points_accessor.byteOffset: points_buffer_view.byteOffset
                                      + points_buffer_view.byteLength
        ],
        dtype="float32",
        count=points_accessor.count * 3,
    )

    color_accessor = gltf.accessors[gltf.meshes[0].primitives[0].attributes.COLOR_0]
    color_buffer_view = gltf.bufferViews[color_accessor.bufferView]
    colors = np.frombuffer(
        binary_blob[
        color_buffer_view.byteOffset
        + color_accessor.byteOffset: color_buffer_view.byteOffset
                                     + color_buffer_view.byteLength
        ],
        dtype="float32",
        count=color_accessor.count * 3,
    )

    points = points.reshape((-1, 3))
    points = points.astype('float64')
    colors = colors.reshape((-1, 3))

    red = np.round(np.power(np.array(colors[:, 0]), 1 / 2.2) * 255).astype('uint8')
    green = np.round(np.power(np.array(colors[:, 1]), 1 / 2.2) * 255).astype('uint8')
    blue = np.round(np.power(np.array(colors[:, 2]), 1 / 2.2) * 255).astype('uint8')

    return {'x': np.array(points[:, 0]) + translation[0], 'y': np.array(points[:, 1]) + translation[1],
            'z': np.array(points[:, 2]) + translation[2], 'red': red, 'green': green, 'blue': blue}



def recurse(f, body, listAll, extraKey = None):

    r = f(body)
    if listAll:
        nextPageStart = r['nextPageStart']
        while nextPageStart != None:
            body['pageStart'] = nextPageStart
            r_new = f(body)
            nextPageStart = r_new['nextPageStart']
            if 'size' in r.keys():
                r['size'] = r['size'] + r_new['size']
            if extraKey == None:
                r['result'] =  r['result'] + r_new['result']
            else:
                r['result'][extraKey] = r['result'][extraKey] +  r_new['result'][extraKey]
                
        r['nextPageStart'] = None
    return r


def stringToDate(date):
    date = sanitize.validString('date', date, True)

    try:
        d = datetime.strptime(date, "%Y-%m-%dT%H:%M:%S.%fZ")
    except:
        try:
            d = datetime.strptime(date, "%Y-%m-%d %H:%M:%S.%f")
        except:
            d = datetime.strptime(date, "%Y-%m-%d %H:%M:%S")

    return d

def dateToString(date):
    date = sanitize.validDate('date', date , True)
    
    d = date.strftime(date, "%Y-%m-%d %H:%M:%S.%fZ")
    return d

def plotRaster(raster):
    raster = sanitize.validNumpyArray('raster', raster, True)

    if len(raster.shape) != 3 and len(raster.shape) != 2:
        raise ValueError('raster must have 2 or 3 dimensions')

    if len(raster.shape) == 2:
        raster = np.expand_dims(raster, 0)

    

    if raster.shape[0] ==3 or raster.shape[0] ==4:
        raster = raster[0:3,:,:]
        raster = np.transpose(raster, [1,2,0])
        minimum = np.min(raster)
        maximum = np.max(raster)
        if minimum == maximum:
            maximum = minimum + 1
        raster = raster - minimum
        raster = raster / (maximum-minimum)
        raster = raster * 255
        image = Image.fromarray(raster.astype('uint8'))
        image.show()

    else:
        plt.imshow(raster[0,:,:], interpolation='none')
        plt.show()
    
def plotVector(features):
    features = sanitize.validGeopandas('features', features, True)
    features.plot()


def chunks(l, n = 3000):
    l = sanitize.validList('l', l, True)
    n = sanitize.validInt('n', n , False)
    result = list()
    for i in range(0, len(l), n):
        result.append(l[i:i+n])
    return(result)
    

 
def cover(bounds, width):
    w = width
    w = sanitize.validInt('w',w, True)
    bounds = sanitize.validBounds('bounds',bounds, True)

    x1 = bounds['xMin']
    y1 = bounds['yMin']
    x2 = bounds['xMax']
    y2  = bounds['yMax']

    step_y =  w/geodesic((y1,x1), (y1 - 1,x1)).meters
    parts_y = math.floor((y2 - y1)/ step_y + 1)

    y1_vec = y1 + np.arange(0, parts_y )*step_y
    y2_vec = y1 + np.arange(1, parts_y +1 )*step_y
        
    steps_x = [   w/geodesic((y,x1), (y,x1+1)).meters for y in y1_vec  ]

    parts_x = [math.floor( (x2-x1) /step +1 ) for step in steps_x ]      
        

    coords = pd.DataFrame()
    for n in np.arange(len(parts_x)):
        x1_sq = [ x1 + j*steps_x[n] for j in np.arange(0,parts_x[n]) ]
        x2_sq = [ x1 + j*steps_x[n] for j in np.arange(1, parts_x[n]+1) ]
        coords_temp = {'x1': x1_sq, 'x2': x2_sq, 'y1': y1_vec[n], 'y2':y2_vec[n]}
        coords = coords.append(pd.DataFrame(coords_temp))

    cover = [Polygon([ (coords['x1'].iloc[j] , coords['y1'].iloc[j]) , (coords['x2'].iloc[j] , coords['y1'].iloc[j]), (coords['x2'].iloc[j] , coords['y2'].iloc[j]), (coords['x1'].iloc[j] , coords['y2'].iloc[j]) ]) for j in np.arange(coords.shape[0])]
     


    coords = gpd.GeoDataFrame({'geometry': cover, 'x1':coords['x1'], 'x2':coords['x2'], 'y1':coords['y1'], 'y2':coords['y2'] })

    coords.crs = 'epsg:4326'

    return(coords)
    

    
def loadingBar(count,total):
    
    count = sanitize.validInt('count', count, True)
    total = sanitize.validInt('total', total, True)
    
    if total == 0:
        return
    else:
        percent = float(count)/float(total)*100
        sys.stdout.write("\r" + str(int(count)).rjust(3,'0')+"/"+str(int(total)).rjust(3,'0') + ' [' + '='*int(percent) + ' '*(100-int(percent)) + ']')


def reprojectRaster(r, sourceExtent, targetExtent, targetWidth, targetHeight, sourceEpsg, targetEpsg, interpolation = 'nearest'):
    
    sanitize.validNumpyArray('r', r, True)
    targetExtent, targetCrs = sanitize.validBoundsCrsCombination(['targetExtent', 'targetEpsg'], [targetExtent, 'EPSG:' + str(targetEpsg)], True)
    sourceExtent, sourceCrs = sanitize.validBoundsCrsCombination(['sourceExtent', 'sourceEpsg'], [sourceExtent, 'EPSG:' + str(sourceEpsg)], True)
    targetWidth = sanitize.validInt('targetWidth', targetWidth, True)
    targetHeight = sanitize.validInt('targetHeight', targetHeight, True)
    interpolation = sanitize.validString('interpolation', interpolation, True)
    
    if not interpolation in ['nearest', 'bilinear']:
        raise ValueError('interpolation must be one of nearest or bilinear')
    
    if len(r.shape) !=3:
        raise ValueError('r must be 3 dimensional')
    if interpolation != 'nearest' and interpolation != 'linear':
        raise ValueError('interpolation must be either nearest or linear')
        
        

    src_transform = rasterio.transform.from_bounds(sourceExtent['xMin'], sourceExtent['yMin'], sourceExtent['xMax'], sourceExtent['yMax'], r.shape[2], r.shape[1])
    dst_transform = rasterio.transform.from_bounds(targetExtent['xMin'], targetExtent['yMin'], targetExtent['xMax'], targetExtent['yMax'], targetWidth, targetHeight)
    destination = np.zeros((r.shape[0],targetHeight,targetWidth))

    if interpolation == 'nearest':
        interpolation = Resampling.nearest        
    else:
        interpolation = Resampling.bilinear

    for i in np.arange(r.shape[0]):
        warp(
            r[i,:,:],
            destination[i,:,:],
            src_transform=src_transform,
            src_crs=sourceCrs,
            dst_transform=dst_transform,
            dst_crs=targetCrs,
            resampling=interpolation)
    

    return {'raster':destination, 'transform':dst_transform, 'extent':targetExtent, 'epsg':targetEpsg}


def swapXY(extent):
    return {'xMin': extent['yMin'], 'xMax':extent['yMax'], 'yMin':extent['xMin'], 'yMax':extent['xMax']}


def transformPoint( point, sourceEpsg, targetEpsg):
    sourceCrs = 'EPSG:' + str(sourceEpsg)
    targetCrs = 'EPSG:' + str(targetEpsg)

    try:
        point = geometry.Point(point)
    except:
        raise ValueError('point must be a tuple with two floats')
    df = gpd.GeoDataFrame({'geometry':[point]})    
    try:
        df.crs = sourceCrs
    except:
        raise ValueError('sourceCrs not a valid crs')
    try:
        df = df.to_crs(targetCrs)
    except:
        raise ValueError('targetCrs not a valid crs')
    x = df.bounds['minx'].values[0]
    y = df.bounds['miny'].values[0]

    return(x,y)

#funciton to store image given bounds, and raster and crs
def saveRaster(targetFile, r, epsg, extent = None, transform = None):
    
    if type(extent) == type(None) and type(transform) == type(None):
        raise ValueError('You must provide either an extent or a transform')
        
    crs = 'EPSG:' + str(epsg)
    if type(extent) != type(None):
        sanitize.validBoundsCrsCombination([extent, crs], [extent, crs], True)
        
    r = sanitize.validNumpyArray('r', r, True)
    
    if len(r.shape) != 3:
        raise ValueError('r must be 3 dimensional')

    w = r.shape[2]
    h = r.shape[1]
    if transform == None:
        transform = rasterio.transform.from_bounds(extent['xMin'], extent['yMin'], extent['xMax'], extent['yMax'], w, h)
    con = rasterio.open(targetFile , 'w', driver='Gtiff',compress="lzw",
                    height = h, width = w,
                    count= r.shape[0], dtype=r.dtype,
                    crs=  crs ,
                    transform=transform)
    
    con.write(r)
    con.close()

def saveVector(targetFile, features):
    features.to_file(targetFile)


q_running = multiprocessing.Queue()

def cutIntoTiles(features, zoom, cores = 1, buffer = 0):
    features, bounds = reprojectWithBounds(sh = features, targetCrs = 'EPSG:3857', cores= cores)
    features['geometryId'] = [str(uuid.uuid4()) for x in np.arange(features.shape[0])]
    features['tileId'] = str(uuid.uuid4())
    count  = features.shape[0]
    
    types = np.unique(np.array([str(type(x)) for x in features['geometry'].values]))
    chosen = None
    for t in types:
        if 'collection' in t:
            raise ValueError('geometry collections are not allowed')
        if 'polygon' in t:
            newChosen = 'polygon'
            if chosen != None and chosen !=newChosen:
                raise ValueError('Geopandas cannot contain mixed geometry types, use either lines, points of polygons.')
            else:
                chosen = 'polygon'
        if 'line' in t:
            newChosen = 'line'
            if chosen != None and chosen !=newChosen:
                raise ValueError('Geopandas cannot contain mixed geometry types, use either lines, points of polygons.')
            else:
                chosen = 'line'
        if 'point' in t:
            newChosen = 'point'
            if chosen != None and chosen !=newChosen:
                raise ValueError('Geopandas cannot contain mixed geometry types, use either lines, points of polygons.')
            else:
                chosen = 'point'

    

    LEN = 2.003751e+07
    tile = {'xMin' : -LEN, 'xMax': LEN, 'yMin':-LEN, 'yMax':LEN} 
    args = (tile, 0, zoom, features, bounds, cores, buffer, count, chosen)
    sh_end = manageTile(args)
    return sh_end
    
    

def manageTile(args):
    tile, depth, zoom, sh, bounds, cores, buffer, count, chosen = args
    if zoom == depth:
        return sh
    if depth == zoom -1:
        t_buffer = buffer
    else:
        t_buffer = 0


    newTiles = splitTile(tile, t_buffer)
    shs = []
    tile = newTiles[0]
    args = []

    maxCount = 0
    for tile in newTiles:        
        sh_new, bounds_new = cut(sh, bounds, tile, chosen)
        if sh_new.shape[0] > maxCount:
            maxCount = sh_new.shape[0]
        if sh_new.shape[0] > 0:
            args = args + [(tile, depth+1, zoom, sh_new, bounds_new, cores, buffer, count, chosen,)]
    
    
    if q_running.qsize() < cores -1 and  count * 0.2 > maxCount :
        q_running.put('x')        
        with Pool(4) as p:
            shs = p.map(manageTile, args)
        shs = list(shs)
        q_running.get()
    else:
        for arg in args:
            sh_new = manageTile(arg)
            shs = shs + [sh_new]
        
        
    sh_total = pd.concat(shs)
    return sh_total


def cut(sh, bounds, tile, chosen):
        #remove things without
        isWithout = np.logical_or( np.logical_or(bounds['minx'].values > tile['xMax'],bounds['maxx'].values < tile['xMin'] ), np.logical_or(bounds['miny'].values > tile['yMax'],bounds['maxy'].values < tile['yMin'] ))
                
        sh_inside = sh[ np.logical_not( isWithout) ]
        bounds_inside = bounds[np.logical_not( isWithout)]
        
        #remove things within
        isWithinX = np.logical_and( np.logical_and(bounds_inside['minx'].values < tile['xMax'],bounds_inside['minx'].values > tile['xMin'] ), np.logical_and(bounds_inside['maxx'].values < tile['xMax'],bounds_inside['maxx'].values > tile['xMin'] ))
        isWithinY = np.logical_and( np.logical_and(bounds_inside['miny'].values < tile['yMax'],bounds_inside['miny'].values > tile['yMin'] ), np.logical_and(bounds_inside['maxy'].values < tile['yMax'],bounds_inside['maxy'].values > tile['yMin'] ))
        isWithin = np.logical_and(isWithinX, isWithinY)

        sh_within = sh_inside[isWithin]        
        bounds_within = bounds_inside[isWithin]
        
        #now intersect
        sh_intersects = sh_inside[np.logical_not(isWithin)]
        poly = tileToPolygon(tile)
        sh_intersects['geometry'] = sh_intersects.intersection(poly)
        
        #dissolve the geometry collections
        if sh_intersects.shape[0] >0:
            collections = np.array(['collection' in str(type(x)) for x in sh_intersects['geometry']])
            geometryCollections = sh_intersects['geometry'][collections].values
    
            newGeometries = [ cascaded_union([geometry for geometry in collection if chosen in str(type(geometry))]) for collection in geometryCollections   ]
    
            sh_intersects['geometry'][collections] = newGeometries
            
            correct_type = [chosen in str(type(x)) for x in sh_intersects['geometry'].values]
            sh_intersects = sh_intersects[correct_type]
            sh_intersects = sh_intersects[np.logical_not(sh_intersects.is_empty)]
            sh_intersects = sh_intersects[sh_intersects.is_valid]

        
        bounds_intersects = sh_intersects.bounds

        sh_result = pd.concat([sh_intersects, sh_within])
        sh_result['tileId'] = str(uuid.uuid4())
        bounds_result = pd.concat([bounds_intersects, bounds_within])


        return sh_result, bounds_result

def tileToPolygon(tile):
    return geometry.Polygon([ (tile['xMin'], tile['yMin']), (tile['xMin'], tile['yMax']), (tile['xMax'], tile['yMax']), (tile['xMax'], tile['yMin']) ])    
    

def splitTile(tile, buffer):
    
    b = buffer * (tile['xMax'] - tile['xMin'])
    xMiddle = tile['xMin'] + (tile['xMax'] - tile['xMin'])/2
    yMiddle = tile['yMin'] + (tile['yMax'] - tile['yMin'])/2
    yMin = tile['yMin']
    xMin = tile['xMin']
    yMax = tile['yMax']
    xMax = tile['xMax']
    
    
    T1 = {'xMin': xMin -b, 'xMax':xMiddle+b, 'yMin': yMin-b, 'yMax': yMiddle+b  }
    T2 = {'xMin': xMin-b, 'xMax':xMiddle+b, 'yMin': yMiddle-b, 'yMax': yMax+b  }
    
    T3 = {'xMin': xMiddle-b, 'xMax':xMax+b, 'yMin': yMin-b, 'yMax': yMiddle+b  }
    T4 = {'xMin': xMiddle-b, 'xMax':xMax+b, 'yMin': yMiddle-b, 'yMax': yMax+b  }
    
    
    return [T1, T2, T3, T4]

    

def reprojectVector(features, targetEpsg, cores = 1):
    sh = features
    targetCrs = 'EPSG:' + str(targetEpsg)
    shs = np.array_split(sh, cores)
    args = list(zip(shs, np.repeat(targetCrs, cores) ))
    with Pool(cores) as p:
        shs = p.map(reprojectSub, args )
    
    sh = pd.concat([x[0] for x in shs])
    return sh

def reprojectWithBounds(sh, targetCrs, cores = 1):

    N = max(round(sh.shape[0]/ 50000),1)
    
    shs_1 = np.array_split(sh, N)
    shs_total = []
    for sh in shs_1:
        shs = np.array_split(sh, cores)
        args = list(zip(shs, np.repeat(targetCrs, cores) ))
        with Pool(cores) as p:
            shs = p.map(reprojectSub, args )
        shs = list(shs)
        shs_total = shs_total + shs
        
    sh = pd.concat([x[0] for x in shs_total])
    bounds = pd.concat([x[1] for x in shs_total])
    return sh, bounds


def reprojectSub(args):
    sh = args[0]
    targetCrs = args[1]
    try:
        sh = sh.to_crs(str(targetCrs))
    except:
        sh = sh.to_crs(targetCrs.replace('EPSG:', ''))


    
    is_poly = ['poly' in str(type(x)) for x in sh['geometry'].values]
    if np.sum(is_poly) > 0:
        sh['geometry'][is_poly] = sh[is_poly].buffer(0)    
    
    bounds = sh.bounds
    return sh, bounds



def getActualExtent(minx, maxx, miny, maxy, crs, out_crs = 3857):
    
    LEN = 2.003751e+07
    STEPS = 10
    
    x_step = (maxx - minx) / STEPS
    y_step = (maxy - miny) / STEPS
    
    points = [ geometry.Point((minx + (i) * x_step, miny + (j) * y_step))  for i in np.arange(STEPS+1) for j in np.arange(STEPS+1)]
    df = gpd.GeoDataFrame({'geometry':points})
    try:
        df.crs =  crs
    except:
        return {'status': 400, 'message': 'Invalid epsg code'}
        
    try:
        df_wgs = df.to_crs('EPSG:' + str(out_crs))
    except:
        return {'status': 400, 'message': 'Invalid extent and epsg combination'}


    #in case points fall outside defined area we restrict to the -85, 85, 180, -180 region
    xs = df_wgs.bounds['minx'].values
    ys = df_wgs.bounds['miny'].values
    xs[xs == np.inf] = LEN    
    xs[xs == -np.inf] = -LEN    
    ys[ys == np.inf] = LEN    
    ys[ys == -np.inf] = -LEN    
    
    minX = np.min(xs)
    maxX = np.max(xs)
    minY = np.min(ys)
    maxY = np.max(ys)

    minX = max(-LEN, minX)
    maxX = min(LEN, maxX)
    minY = max(-LEN, minY)
    maxY = min(LEN, maxY)
    
    
    
    return {'status': 200, 'message': {'xMin':minX, 'xMax': maxX, 'yMin': minY, 'yMax':maxY}}

def mergeGeometriesByColumn(features, columnName, cores = 1):
    names = np.unique(features[columnName].values)
    N = math.floor(len(names)/cores) + 1
    names_chunks = chunks(names, N)
    args = [{'features': features, 'columnName':columnName, 'names':names_cunk} for names_cunk in names_chunks]
    if cores > 1:
        with Pool(4) as p:
            shs = p.map(mergeGeometriesByColumnPart, args)
        shs = list(shs)
        return pd.concat(shs)

    else:
        return mergeGeometriesByColumnPart( {'features':features, 'columnName':columnName, 'names':names})
    


def mergeGeometriesByColumnPart(args):

    features = args['features']
    columnName = args['columnName']
    names = args['names']
    
    sh_new = []
    for i in np.arange(len(names)):
        name = names[i]
        sh_name = features[features[columnName] == name ]
        sh_name['geometry'] = sh_name.unary_union
        sh_new = sh_new + [sh_name[0:1]]
    sh_new = pd.concat(sh_new)
    return sh_new
        



def simplifyGeometries(features, tolerance, preserveTopology=True, removeIslands = True, cores = 1):
    
    shs = np.array_split(features, cores)
        
    args = [ {'sh': shs[i], 'tolerance':tolerance, 'preserveTopology': preserveTopology, 'removeIslands':removeIslands} for i in np.arange(len(shs))]
    if cores >1:
        with Pool(cores) as p:
            shs = p.map(simplifySub, args)    
            result = pd.concat(list(shs))
    else:
        result = simplifySub(args[0])
    return result
    
    
def simplifySub(args):
    sh_sub = args['sh']
    tolerance = args['tolerance']
    preserveTopology = args['preserveTopology']
    removeIslands = args['removeIslands']

    fractionArea = tolerance**2

    if removeIslands:
        multipolygons = sh_sub['geometry'][sh_sub['geometry'].type == 'MultiPolygon' ]
        newMultiPolygons = [shapely.ops.unary_union([x.buffer(0) for x in multipolygons.values[i] if x.area > fractionArea ]) if len([x for x in multipolygons.values[i] if x.area > fractionArea ]) > 0 else multipolygons.values[i]  for i in np.arange( len(multipolygons))]
        
        #newMultiPolygons = [x if len(x) > 1 else x[0] for x in newMultiPolygons]
        sh_sub['geometry'][sh_sub['geometry'].type == 'MultiPolygon' ] = newMultiPolygons


    sh_sub['geometry'] = sh_sub.simplify(tolerance, preserve_topology=preserveTopology)
    is_poly = ['poly' in str(type(x)) for x in sh_sub['geometry'].values]
    sh_sub['geometry'][is_poly] = sh_sub[is_poly].buffer(0)

    return sh_sub
    
    



from ellipsis.path.raster.timestamp.root import add
from ellipsis.path.raster.timestamp.root import edit
from ellipsis.path.raster.timestamp.root import delete
from ellipsis.path.raster.timestamp.root import activate
from ellipsis.path.raster.timestamp.root import deactivate
from ellipsis.path.raster.timestamp.root import getBounds
from ellipsis.path.raster.timestamp.root import analyse
from ellipsis.path.raster.timestamp.root import getRaster
from ellipsis.path.raster.timestamp.root import getDownsampledRaster
from ellipsis.path.raster.timestamp.root import trash
from ellipsis.path.raster.timestamp.root import recover
from ellipsis.path.raster.timestamp.root import getSampledRaster
from ellipsis.path.raster.timestamp.root import getValuesAlongLine
from ellipsis.path.raster.timestamp.root import contour

from ellipsis.path.raster.timestamp import file
from ellipsis.path.raster.timestamp import order





from ellipsis import apiManager
from ellipsis import sanitize



def add(pathId, name, method, parameters, token, default = True):
    pathId = sanitize.validUuid('pathId', pathId, True) 
    method = sanitize.validString('method', method, True)
    parameters = sanitize.validObject('parameters', parameters, True)
    token = sanitize.validString('token', token, True)
    default = sanitize.validBool('default', default, True)

    body = {'name': name, 'default':default, 'method': method, 'parameters':parameters}

    r = apiManager.post('/path/' + pathId + '/vector/style', body, token)
    return r


def edit(pathId, styleId, token, name = None, method = None, parameters = None, default = None):
    pathId = sanitize.validUuid('pathId', pathId, True) 
    styleId = sanitize.validUuid('styleId', styleId, True) 
    method = sanitize.validString('method', method, False)
    parameters = sanitize.validObject('parameters', parameters, False)
    token = sanitize.validString('token', token, True)
    default = sanitize.validBool('default', default, False)
    name = sanitize.validString('name', name, False)

    body = {'name': name, 'default':default, 'method': method, 'parameters':parameters}

    r = apiManager.patch('/path/' + pathId + '/vector/style/' + styleId, body, token)
    return r
    
    
def delete(pathId, styleId, token):
    pathId = sanitize.validUuid('pathId', pathId, True) 
    styleId = sanitize.validUuid('styleId', styleId, True) 
    token = sanitize.validString('token', token, True)
    

    r = apiManager.delete('/path/' + pathId + '/vector/style/' + styleId, None, token)
    return r
    


